/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

///////////////////////////////////////////////////////////////////////////
/// @file	sys_win.c
/// @brief	Win32 system interface code
/// @defgroup system_functions System Functions
///////////////////////////////////////////////////////////////////////////

#include "quakedef.h"
#include "winquake.h"
#include "errno.h"
#include "resource.h"
#include "conproc.h"

#define MINIMUM_WIN_MEMORY		0x0880000	///< Minimum needed RAM (?): 8.5 MiB
#define MAXIMUM_WIN_MEMORY		0x1000000	///< Maximum needed RAM (?): 16 MiB

#define CONSOLE_ERROR_TIMEOUT	60.0		///< Number of seconds to wait on Sys_Error running.
//  dedicated before exiting
#define PAUSE_SLEEP		50					///< Sleep time on pause or minimization.
#define NOT_FOCUS_SLEEP	20					///< Sleep time when not focus.	

int		starttime;							///< Unused?
bool	ActiveApp, Minimized;				///< Used to stop the game from running if it's not selected or when it's minimized.
bool	WinNT;

static double	pfreq;						///< Performance Frequency.
static double	curtime = 0.0;				///< Game Time.
static double	lastcurtime = 0.0;			///< Holds the current time from the last loop.
static int		lowshift;					///< Used for the performance counter
bool			isDedicated;				///< Is the game a dedicated server?
static bool		sc_return_on_enter = false;	///< Used to stop the dedicated server (?).
HANDLE			hinput, houtput;			///< Input and output handlers

static char		*tracking_tag = "Clams & Mooses";	///< Unused?

static HANDLE	tevent;						///< Time(?) Event
static HANDLE	hFile;						///< Used for QHOST to hook into the console
static HANDLE	heventParent;				///< Used for QHOST to hook into the console
static HANDLE	heventChild;				///< Used for QHOST to hook into the console

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Unused function.
///////////////////////////////////////////////////////////////////////////
void MaskExceptions (void);

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Set var lastcurtime to curtime.
///////////////////////////////////////////////////////////////////////////
void Sys_InitFloatTime (void);

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Unused function.
///////////////////////////////////////////////////////////////////////////
void Sys_PushFPCW_SetHigh (void);

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Unused function.
///////////////////////////////////////////////////////////////////////////
void Sys_PopFPCW (void);

volatile int sys_checksum;



///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Note sure what this does, but it has to do with the preallocated
///			memory from WinMain. It seems to make sure that there's 16 MiB
///			available for the game
/// @param	ptr The to "page" memory pointer
/// @param	size The memory block size.
///////////////////////////////////////////////////////////////////////////
void Sys_PageIn (void *ptr, int size)
{
	byte *x;
	int j, m, n;
	
	// touch all the memory to make sure it's there. The 16-page skip is to
	// keep Win 95 from thinking we're trying to page ourselves in (we are
	// doing that, of course, but there's no reason we shouldn't)
	x = (byte *) ptr;
	
	for (n = 0; n < 4; n++)
	{
		for (m = 0; m < (size - 16 * 0x1000); m += 4)
		{
			sys_checksum += *(int *) &x[m];
			sys_checksum += *(int *) &x[m + 16 * 0x1000];
		}
	}
}


///////////////////////////////////////////////////////////////////////////
// FILE IO
///////////////////////////////////////////////////////////////////////////

#define	MAX_HANDLES 10
FILE	*sys_handles[MAX_HANDLES];	///< @b Global. Holds several @e FILE pointers.

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Search the system handles and return the first free handle.
///			If this isn't possible (because there are no more handles) return -1
/// @return The first free file handle.
///////////////////////////////////////////////////////////////////////////
int findhandle (void)
{
	for (int i = 1; i < MAX_HANDLES; i++)
	{
		if (!sys_handles[i])
		{
			return i;
		}
	}

	Sys_Error ("out of handles");
	return -1;
}

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Checks the size (length) of a file.
/// @param	f The file to check for its size.
/// @return	Filesize
///////////////////////////////////////////////////////////////////////////
int filelength (FILE *f)
{
	int pos;	// Begin position of the file.
	int end;	// End position of the file.
	int t;		// 
	
	t = VID_ForceUnlockedAndReturnState ();
	
	pos = ftell (f);
	fseek (f, 0, SEEK_END);
	end = ftell (f);
	fseek (f, pos, SEEK_SET);
	
	VID_ForceLockState (t);
	
	return end;
}

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Open @b file and attach a  @b handle to it
/// @param	path If non-null, full pathname of the file.
/// @param	hndl If non-null, the hndl.
/// @return	The file's length, -1 on error
///////////////////////////////////////////////////////////////////////////
int Sys_FileOpenRead (char *path, int *hndl)
{
	FILE *f;
	int i;
	int retval;
	int t;
	
	t = VID_ForceUnlockedAndReturnState ();
	
	i = findhandle ();
	
	f = fopen (path, "rb");
	
	if (!f)
	{
		*hndl = -1;
		retval = -1;
	}
	else
	{
		sys_handles[i] = f;
		*hndl = i;
		retval = filelength (f);
	}
	
	VID_ForceLockState (t);
	
	return retval;
}

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	System file open write
/// @param	path Filepointer path; If non-null, full pathname of the file.
/// @return	The first free (file) handle
///////////////////////////////////////////////////////////////////////////
int Sys_FileOpenWrite (char *path)
{
	FILE	*f;
	int		i;
	int		t;
	
	t = VID_ForceUnlockedAndReturnState ();
	
	i = findhandle ();
	
	f = fopen (path, "wb");
	if (!f)
		Sys_Error ("Error opening %s: %s", path, strerror (errno));
		
	sys_handles[i] = f;
	
	VID_ForceLockState (t);
	
	return i;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Close the handle-th file in @e sys_handles[]
/// @param	handle Close the nth file.
///////////////////////////////////////////////////////////////////////////
void Sys_FileClose (int handle)
{
	int	t;
	
	t = VID_ForceUnlockedAndReturnState ();
	fclose (sys_handles[handle]);
	sys_handles[handle] = NULL;
	VID_ForceLockState (t);
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Point the nth file in @e sys_handles[] to the positionth byte
/// @param	handle The nth file to change its position pointer.
/// @param	position The position to point to.
///////////////////////////////////////////////////////////////////////////
void Sys_FileSeek (int handle, int position)
{
	int t;
	
	t = VID_ForceUnlockedAndReturnState ();
	fseek (sys_handles[handle], position, SEEK_SET);
	VID_ForceLockState (t);
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Read from the nth file in @e sys_handles[] and return the value of the count position
/// @param	handle The nth file to read
/// @param	dest The mem position to write the read bytes to.
/// @param	count The block size to read.
/// @return	Size of @e count (To check if reading worked)
///////////////////////////////////////////////////////////////////////////
int Sys_FileRead (int handle, void *dest, int count)
{
	int t, x;
	
	t = VID_ForceUnlockedAndReturnState ();
	x = fread (dest, 1, count, sys_handles[handle]);
	VID_ForceLockState (t);
	return x;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Write to the nth file in @e sys_handles[] and return the value of the count position
/// @param	handle The nth file to read
/// @param	data The data to write.
/// @param	count The block size to write.
/// @return	The size of @e count.
///////////////////////////////////////////////////////////////////////////
int Sys_FileWrite (int handle, void *data, int count)
{
	int t, x;
	
	t = VID_ForceUnlockedAndReturnState ();
	x = fwrite (data, 1, count, sys_handles[handle]);
	VID_ForceLockState (t);
	return x;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Check if a file can be opened.
/// @param	path The file to check.
///////////////////////////////////////////////////////////////////////////
int	Sys_FileTime (char *path)
{
	FILE	*f;
	int		t, retval;
	
	t = VID_ForceUnlockedAndReturnState ();
	
	f = fopen (path, "rb");
	
	if (f)
	{
		fclose (f);
		retval = 1;
	}
	else
	{
		retval = -1;
	}
	
	VID_ForceLockState (t);
	return retval;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	
/// @param	path
///////////////////////////////////////////////////////////////////////////
void Sys_mkdir (char *path)
{
	_mkdir (path);
}


///////////////////////////////////////////////////////////////////////////
/// 
/// SYSTEM IO
/// 
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	
/// @param	startaddr
/// @param	length
///////////////////////////////////////////////////////////////////////////
void Sys_MakeCodeWriteable (unsigned long startaddr, unsigned long length)
{
	DWORD  flOldProtect;
	
	if (!VirtualProtect ( (LPVOID) startaddr, length, PAGE_READWRITE, &flOldProtect))
		Sys_Error ("Protection change failed\n");
}


#ifndef _M_IX86

void Sys_SetFPCW (void)
{
}

void Sys_PushFPCW_SetHigh (void)
{
}

void Sys_PopFPCW (void)
{
}

void MaskExceptions (void)
{
}

#endif

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Initialized some system functionality.
/// @todo	List what this function specifically does.
///////////////////////////////////////////////////////////////////////////
void Sys_Init (void)
{
	LARGE_INTEGER PerformanceFreq;
	unsigned int  lowpart, highpart;
	OSVERSIONINFO vinfo;
	
	MaskExceptions ();	// This doesn't DO anything, does it?
	Sys_SetFPCW ();		// ditto with this one...
	
	if (!QueryPerformanceFrequency (&PerformanceFreq))
	{
		Sys_Error ("No hardware timer available");
	}
	
	// get 32 out of the 64 time bits such that we have around
	// 1 microsecond resolution
	lowpart = (unsigned int) PerformanceFreq.LowPart;
	highpart = (unsigned int) PerformanceFreq.HighPart;
	lowshift = 0;
	
	while (highpart || (lowpart > 2000000.0))
	{
		lowshift++;
		lowpart >>= 1;
		lowpart |= (highpart & 1) << 31;
		highpart >>= 1;
	}
	
	pfreq = 1.0 / (double) lowpart;
	
	Sys_InitFloatTime ();
	
	vinfo.dwOSVersionInfoSize = sizeof (vinfo);
	
	if (!GetVersionEx (&vinfo))
	{
		Sys_Error ("Couldn't get OS info");
	}
		
	if ((vinfo.dwMajorVersion < 4) ||
			(vinfo.dwPlatformId == VER_PLATFORM_WIN32s))
	{
		Sys_Error ("WinQuake requires at least Win95 or NT 4.0");
	}
	
	if (vinfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
		WinNT = true;
	else
		WinNT = false;
}

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Print an error message to either the command prompt
///			or a pop up window and exit the program.
///	@param	error One or more error messages
///	@return void
///////////////////////////////////////////////////////////////////////////
void Sys_Error (char *error, ...)
{
	va_list		argptr;
	char		text[1024], text2[1024];
	char		*text3 = "Press Enter to exit\n";
	char		*text4 = "***********************************\n";
	char		*text5 = "\n";
	DWORD		dummy;
	double		starttime;
	static int	in_sys_error0 = 0;
	static int	in_sys_error1 = 0;
	static int	in_sys_error2 = 0;
	static int	in_sys_error3 = 0;
	
	if (!in_sys_error3)
	{
		in_sys_error3 = 1;
		VID_ForceUnlockedAndReturnState ();
	}
	
	va_start (argptr, error);
	vsprintf (text, error, argptr);
	va_end (argptr);
	
	if (isDedicated)
	{
		va_start (argptr, error);
		vsprintf (text, error, argptr);
		va_end (argptr);
		
		sprintf (text2, "ERROR: %s\n", text);
		WriteFile (houtput, text5, strlen (text5), &dummy, NULL);
		WriteFile (houtput, text4, strlen (text4), &dummy, NULL);
		WriteFile (houtput, text2, strlen (text2), &dummy, NULL);
		WriteFile (houtput, text3, strlen (text3), &dummy, NULL);
		WriteFile (houtput, text4, strlen (text4), &dummy, NULL);
		
		
		starttime = Sys_FloatTime ();
		sc_return_on_enter = true;	// so Enter will get us out of here
		
		while (!Sys_ConsoleInput () &&
				( (Sys_FloatTime () - starttime) < CONSOLE_ERROR_TIMEOUT))
		{
		}
	}
	else
	{
		// switch to windowed so the message box is visible, unless we already
		// tried that and failed
		if (!in_sys_error0)
		{
			in_sys_error0 = 1;
			VID_SetDefaultMode ();
			MessageBox (NULL, text, "Quake Error",
						MB_OK | MB_SETFOREGROUND | MB_ICONSTOP);
		}
		else
		{
			MessageBox (NULL, text, "Double Quake Error",
						MB_OK | MB_SETFOREGROUND | MB_ICONSTOP);
		}
	}
	
	if (!in_sys_error1)
	{
		in_sys_error1 = 1;
		Host_Shutdown ();
	}
	
// shut down QHOST hooks if necessary
	if (!in_sys_error2)
	{
		in_sys_error2 = 1;
		DeinitConProc ();
	}
	
	exit (1);
}

//////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Send text to the console.
/// @param	fmt The to send text.
//////////////////////////////////////////////////////////////////////////
void Sys_Printf (char *fmt, ...)
{
	va_list		argptr;
	char		text[1024];
	DWORD		dummy;
	
	if (isDedicated)
	{
		va_start (argptr, fmt);
		vsprintf (text, fmt, argptr);
		va_end (argptr);
		
		WriteFile (houtput, text, strlen (text), &dummy, NULL);
	}
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Shutdown the host, close all file handles, close the console
///			and exit() the program
///////////////////////////////////////////////////////////////////////////
void Sys_Quit (void)
{
	VID_ForceUnlockedAndReturnState ();
	
	Host_Shutdown();
	
	if (tevent)
		CloseHandle (tevent);
		
	if (isDedicated)
		FreeConsole ();
		
// shut down QHOST hooks if necessary
	DeinitConProc ();
	
	exit (0);
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Calculate the current time using static integers and
///			return it as a double.
/// @return	Current @b static @e curtime.
///////////////////////////////////////////////////////////////////////////
double Sys_FloatTime (void)
{
	static int			sametimecount;
	static unsigned int	oldtime;
	static int			first = 1;
	LARGE_INTEGER		PerformanceCount;
	unsigned int		temp, t2;
	double				time;
	
	Sys_PushFPCW_SetHigh ();	// This function doesn't do anthing, does it?
	
	// Retrieves the current value of the high-resolution performance counter.
	QueryPerformanceCounter (&PerformanceCount);
	
	temp = ((unsigned int) PerformanceCount.LowPart >> lowshift) |
		   ((unsigned int) PerformanceCount.HighPart << (32 - lowshift));
	
	// This if part is run once
	if (first)
	{
		oldtime = temp;
		first = 0;
	}
	else
	{
		// check for turnover or backward time
		if ( (temp <= oldtime) && ( (oldtime - temp) < 0x10000000))
		{
			oldtime = temp;	// so we can't get stuck
		}
		else
		{
			t2 = temp - oldtime;
			
			time = (double) t2 * pfreq;
			oldtime = temp;
			
			curtime += time;
			
			if (curtime == lastcurtime)
			{
				sametimecount++;
				
				if (sametimecount > 100000)
				{
					curtime += 1.0;
					sametimecount = 0;
				}
			}
			else
			{
				sametimecount = 0;
			}
			
			lastcurtime = curtime;
		}
	}
	
	Sys_PopFPCW ();		// This function doesn't do anthing, does it?
	
	return curtime;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Set var lastcurtime to curtime.
///////////////////////////////////////////////////////////////////////////
void Sys_InitFloatTime (void)
{
	int	j;
	
	Sys_FloatTime ();
	
	j = COM_CheckParm ("-starttime");
	
	if (j)
	{
		// set specific start time.
		curtime = (double) (Q_atof (com_argv[j+1]));
	}
	else
	{
		curtime = 0.0;
	}
	
	lastcurtime = curtime;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Record console input and record it. (?)
/// @return	@b NULL If the program is dedicated.
/// @return	The console text.
/// @return	@b NULL If there were no console events.
///////////////////////////////////////////////////////////////////////////
char *Sys_ConsoleInput (void)
{
	static char	text[256];		///< Console text (?).
	static int	len;			///< Console text length.
	INPUT_RECORD recs[1024];	///< The last 1024 key inputs (?).
	int count;
	int i, dummy;
	int ch, numread, numevents;
	
	if (!isDedicated)
		return NULL;
		
	for (;;)
	{
		if (!GetNumberOfConsoleInputEvents (hinput, &numevents))
			Sys_Error ("Error getting # of console events");
		
		// No console events
		if (numevents <= 0)
			break;
		
		if (!ReadConsoleInput (hinput, recs, 1, &numread))
			Sys_Error ("Error reading console input");
		
		if (numread != 1)
			Sys_Error ("Couldn't read console input");
		
		// Key has been pressed (?).
		if (recs[0].EventType == KEY_EVENT)
		{
			if (!recs[0].Event.KeyEvent.bKeyDown)
			{
				ch = recs[0].Event.KeyEvent.uChar.AsciiChar;
				
				switch (ch)
				{
					case '\r':
						WriteFile (houtput, "\r\n", 2, &dummy, NULL);
						
						if (len)
						{
							text[len] = 0;
							len = 0;
							return text;
						}
						else if (sc_return_on_enter)
						{
							// special case to allow exiting from the error handler on Enter
							text[0] = '\r';
							len = 0;
							return text;
						}
						
						break;
						
					case '\b':
						WriteFile (houtput, "\b \b", 3, &dummy, NULL);
						if (len)
						{
							len--;
						}
						break;
						
					default:
						if (ch >= ' ')
						{
							WriteFile (houtput, &ch, 1, &dummy, NULL);
							text[len] = ch;
							len = (len + 1) & 0xff;
						}
						
						break;
				}
			}
		}
	}
	
	return NULL;
}


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	Suspends the execution of the current thread until the
///			time-out interval elapses.
///////////////////////////////////////////////////////////////////////////
void Sys_Sleep (void)
{
	// Milliseconds
	Sleep (1);
}

///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief
/// @todo fix brief
///////////////////////////////////////////////////////////////////////////
void Sys_SendKeyEvents (void)
{
	MSG msg;
	
	while (PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
	{
		// we always update if there are any event, even if we're paused
		scr_skipupdate = 0;
		
		if (!GetMessage (&msg, NULL, 0, 0))
			Sys_Quit ();
			
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
}


/*
==============================================================================

 WINDOWS CRAP

==============================================================================
*/


///////////////////////////////////////////////////////////////////////////
/// @ingroup system_functions
/// @brief	
/// @param	time Time in milliseconds to wait until the next check
/// @todo	find out what this actually does, MS docs don't help >:(
///////////////////////////////////////////////////////////////////////////
void SleepUntilInput (int time)
{
	// Waits until one or all of the specified objects are in the signaled
	// state or the time-out interval elapses. The objects can include
	// input event objects, which you specify using the dwWakeMask parameter
	MsgWaitForMultipleObjects (1, &tevent, FALSE, time, QS_ALLINPUT);
}


///////////////////////////////////////////////////////////////////////////
// global accessible variables
///////////////////////////////////////////////////////////////////////////
HINSTANCE	global_hInstance;
int			global_nCmdShow;
char		*argv[MAX_NUM_ARGVS];
static char	*empty_string = "";
HWND		hwnd_dialog;

///////////////////////////////////////////////////////////////////////////
/// @mainpage
/// @b WinMain is the entry point of the program.
/// @see WinMain
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
/// @brief	Entry point of the program.
/// @param	hInstance Full name: Handle Instance.
///			A handle to the current instance of this program.
///	@param	hPrevInstance Handle to the previous instance (unused)
///	@param	lpCmdLine Full name: Long Pointer Command Line.
///			The parameters in the form of one long string
///	@param	nCmdShow Controls how the window is to be shown.
///	@note	Learn about @b Win32 to learn more about handles and shiz.
///			@b Programming @b Windows, @b 5th @b Edition
///			by @i Charles @i Petzold is a good starting point. You @b NEED
///			to know C pretty well though!
///////////////////////////////////////////////////////////////////////////
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG				msg;
	quakeparms_t	parms;			///< Parameters passed to the executable
	double			time, oldtime, newtime;
	MEMORYSTATUS	lpBuffer;
	static	char	cwd[1024];		// Current Working (?) Directory
	int				t;
	RECT			rect;			///< Bounding rectangle of the specified window. Used if program is started without -dedicated
	
	// previous instances do not exist in Win32
	if (hPrevInstance)
	{
		return 0;
	}
	
	// make hInstance and nCmdShow global
	global_hInstance = hInstance;
	global_nCmdShow = nCmdShow;

	lpBuffer.dwLength = sizeof (MEMORYSTATUS);
	GlobalMemoryStatus (&lpBuffer);
	
	if (!GetCurrentDirectory (sizeof (cwd), cwd))
	{
		Sys_Error ("Couldn't determine current directory");
	}
	
	// If cwd's last char == /, remove it
	if (cwd[Q_strlen (cwd)-1] == '/')
	{
		cwd[Q_strlen (cwd)-1] = 0;
	}
	
	parms.basedir = cwd;
	parms.cachedir = NULL;
	
	parms.argc = 1;
	argv[0] = empty_string;
	
	// convert lpCmdLine to parms
	while (*lpCmdLine && (parms.argc < MAX_NUM_ARGVS))
	{
		// while lpCmdLine is outside the readable ANSI chars.
		while (*lpCmdLine && ((*lpCmdLine <= ' ') || (*lpCmdLine > '~')))
			lpCmdLine++;
		
		if (*lpCmdLine)
		{
			argv[parms.argc] = lpCmdLine;
			parms.argc++;
			
			// while lpCmdLine is inside the readable ANSI chars.
			while (*lpCmdLine && ( (*lpCmdLine > ' ') && (*lpCmdLine <= '~')))
				lpCmdLine++;
				
			if (*lpCmdLine)
			{
				// set last char to NULL
				*lpCmdLine = 0;
				lpCmdLine++;	// why this increment?
			}
		}
	}
	
	// point parms to argv
	parms.argv = argv;
	
	COM_InitArgv (parms.argc, parms.argv);
	
	parms.argc = com_argc;
	parms.argv = com_argv;
	
	isDedicated = (COM_CheckParm ("-dedicated") != 0);
	
	if (!isDedicated)
	{
		hwnd_dialog = CreateDialog (hInstance, MAKEINTRESOURCE (IDD_DIALOG1), NULL, NULL);
		
		if (hwnd_dialog)
		{
			if (GetWindowRect (hwnd_dialog, &rect))
			{
				if (rect.left > (rect.top * 2))
				{
					SetWindowPos (hwnd_dialog, 0,
								  (rect.left / 2) - ((rect.right - rect.left) / 2),
								  rect.top, 0, 0,
								  SWP_NOZORDER | SWP_NOSIZE);
				}
			}
			
			ShowWindow (hwnd_dialog, SW_SHOWDEFAULT);
			UpdateWindow (hwnd_dialog);
			SetForegroundWindow (hwnd_dialog);
		}
	}
	
	// take the greater of all the available memory or half the total memory,
	// but at least 8 Mb and no more than 16 Mb, unless they explicitly
	// request otherwise
	parms.memsize = lpBuffer.dwAvailPhys;
	
	if (parms.memsize < MINIMUM_WIN_MEMORY)
		parms.memsize = MINIMUM_WIN_MEMORY;
		
	if (parms.memsize < (lpBuffer.dwTotalPhys >> 1))
		parms.memsize = lpBuffer.dwTotalPhys >> 1;
		
	if (parms.memsize > MAXIMUM_WIN_MEMORY)
		parms.memsize = MAXIMUM_WIN_MEMORY;
	
	// If specified, Quake will start with more memory usage.
	if (COM_CheckParm ("-heapsize"))
	{
		// select the vector after -heapsize
		t = COM_CheckParm ("-heapsize") + 1;
		
		if (t < com_argc)
		{
			parms.memsize = Q_atoi (com_argv[t]) * 1024; // * 1 MiB
		}
	}
	
	parms.membase = malloc (parms.memsize);
	
	if (!parms.membase)
	{
		Sys_Error ("Not enough memory free; check disk space\n");
	}
	
	Sys_PageIn (parms.membase, parms.memsize);
	
	tevent = CreateEvent (NULL, FALSE, FALSE, NULL);
	
	if (!tevent)
		Sys_Error ("Couldn't create event");
		
	if (isDedicated)
	{
		if (!AllocConsole ())
			Sys_Error ("Couldn't create dedicated server console");
		
		hinput = GetStdHandle (STD_INPUT_HANDLE);
		houtput = GetStdHandle (STD_OUTPUT_HANDLE);
		
		// give QHOST a chance to hook into the console
		if ((t = COM_CheckParm ("-HFILE")) > 0)
		{
			if (t < com_argc)
				hFile = (HANDLE) Q_atoi (com_argv[t+1]);
		}
		
		if ((t = COM_CheckParm ("-HPARENT")) > 0)
		{
			if (t < com_argc)
				heventParent = (HANDLE) Q_atoi (com_argv[t+1]);
		}
		
		if ((t = COM_CheckParm ("-HCHILD")) > 0)
		{
			if (t < com_argc)
				heventChild = (HANDLE) Q_atoi (com_argv[t+1]);
		}
		
		InitConProc (hFile, heventParent, heventChild);
	}
	
	Sys_Init();
	
	// because sound is off until we become active
	S_BlockSound();
	
	Sys_Printf ("Host_Init\n");
	/// @todo Work here further.
	Host_Init (&parms);
	
	oldtime = Sys_FloatTime();
	
	///////////////////////////////////////////////////////////////////////
	// MAIN WINDOW MESSAGE LOOP
	///////////////////////////////////////////////////////////////////////
	while (1)
	{
		if (isDedicated)
		{
			newtime = Sys_FloatTime();
			time = newtime - oldtime;
			
			while (time < sys_ticrate.value)
			{
				Sys_Sleep();
				newtime = Sys_FloatTime();
				time = newtime - oldtime;
			}
		}
		else
		{
			// yield the CPU for a little while when paused, minimized, or not the focus
			if ((cl.paused && (!ActiveApp && !DDActive)) || Minimized || block_drawing)
			{
				SleepUntilInput (PAUSE_SLEEP);
				scr_skipupdate = 1;		// no point in bothering to draw
			}
			else if (!ActiveApp && !DDActive)
			{
				SleepUntilInput (NOT_FOCUS_SLEEP);
			}
			
			newtime = Sys_FloatTime();
			time = newtime - oldtime;
		}
		
		Host_Frame (time);
		oldtime = newtime;
	}
	
	/* return success of application */
	return TRUE;
}

